using System;
using System.IO;
using System.Text;


/*
 * Copyright 2007 Lee Glasser
 * 
 * This code is distributed under the terms of the GNU General Public License
 * (GPL).
 * 
 * This file is part of the Includes package.
 * 
 * "Includes.FindReplace" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 * 
 * "Includes.FindReplace" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with "Includes.FindReplace."  If not, see <http://www.gnu.org/licenses/>.
 */


namespace FindReplace
{
    /// <summary>
    /// This class is for searching for byte sequences (case sensitive) in large files (not to exceed 2 GiB).  It utilizes a rolling buffer.
    /// </summary>
    public class LargeFile
    {
        private string _Path;

        /// <summary>
        /// Creates a new instance of LargeFile where the source (path) is used over and over.
        /// </summary>
        /// <param name="path">The source file.</param>
        public LargeFile(string path)
        {
            _Path = path;
        }

        /// <summary>
        /// Simply gets all instances of "find" in the file.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <returns>The offsets of all instances of "find."</returns>
        public int[] Find(string find)
        {
            return Search.Find(Encoding.ASCII.GetBytes(find), _Path, 4096, -1);
        }
        /// <summary>
        /// Gets the "limit" of instances of "find" in the file.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>The offsets of all instances of "find."</returns>
        public int[] Find(string find, int buffersize, int limit)
        {
            return Search.Find(Encoding.ASCII.GetBytes(find), _Path, buffersize, limit);
        }

        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  This is all done in the same file (will overwrite).
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(string find, string replace)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), 4096, -1);
        }
        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(string find, string replace, string destination)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), _Path, destination, 4096, -1);
        }
        /// <summary>
        /// Finds all instances of "find" in "source" and replaces them with "replace."   Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="source">What file to find in.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <returns>Total number replaced.</returns>
        public static int FindReplace(string find, string replace, string source, string destination)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), source, destination, 4096, -1);
        }

        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  This is all done in the same file (will overwrite).
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(string find, string replace, int buffersize, int limit)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), buffersize, limit);
        }
        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(string find, string replace, string destination, int buffersize, int limit)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), _Path, destination, buffersize, limit);
        }
        /// <summary>
        /// Finds all instances of "find" in "source" and replaces them with "replace."   Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="source">What file to find in.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public static int FindReplace(string find, string replace, string source, string destination, int buffersize, int limit)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), source, destination, buffersize, limit);
        }

        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  This is all done in the same file (will overwrite).
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(byte[] find, byte[] replace)
        {
            return FindReplace(find, replace, 4096, -1);
        }
        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(byte[] find, byte[] replace, string destination)
        {
            return FindReplace(find, replace, _Path, destination, 4096, -1);
        }
        /// <summary>
        /// Finds all instances of "find" in "source" and replaces them with "replace."   Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="source">What file to find in.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <returns>Total number replaced.</returns>
        public static int FindReplace(byte[] find, byte[] replace, string source, string destination)
        {
            return FindReplace(find, replace, source, destination, 4096, -1);
        }

        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  This is all done in the same file (will overwrite).
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(byte[] find, byte[] replace, int buffersize, int limit)
        {
            int[] offsets = Search.Find(find, _Path, buffersize, limit);

            FileStream fs = new FileStream(_Path, FileMode.OpenOrCreate);
            BinaryWriter bw = new BinaryWriter(fs);

            int i;
            for (i = 0; i < offsets.Length; i++)
            {
                fs.Position = offsets[i] - 1;
                bw.Write(replace);
            }
            bw.Close();
            fs.Close();
            return i;
        }
        /// <summary>
        /// Finds all instances of "find" and replaces them with "replace."  Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(byte[] find, byte[] replace, string destination, int buffersize, int limit)
        {
            return FindReplace(find, replace, _Path, destination, buffersize, limit);
        }
        /// <summary>
        /// Finds all instances of "find" in "source" and replaces them with "replace."   Saves changes to "destination."
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="source">What file to find in.</param>
        /// <param name="destination">Where to save the modified file.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public static int FindReplace(byte[] find, byte[] replace, string source, string destination, int buffersize, int limit)
        {
            File.Copy(source, destination, true);

            int[] offsets = Search.Find(find, destination, buffersize, limit);

            FileStream fs = new FileStream(destination, FileMode.OpenOrCreate);
            BinaryWriter bw = new BinaryWriter(fs);

            int i;
            for (i = 0; i < offsets.Length; i++)
            {
                fs.Position = offsets[i] - 1;
                bw.Write(replace);
            }
            bw.Close();
            fs.Close();
            return i;
        }
    }
}
