using FindReplace.Types;
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;


/*
 * Copyright 2007 Lee Glasser
 * 
 * This code is distributed under the terms of the GNU General Public License
 * (GPL).
 * 
 * This file is part of the Includes package.
 * 
 * "Includes.FindReplace" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 * 
 * "Includes.FindReplace" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with "Includes.FindReplace."  If not, see <http://www.gnu.org/licenses/>.
 */


namespace FindReplace
{
    /// <summary>
    /// Contains all the core searching algorithms.
    /// </summary>
    public static class Search
    {
        /// <summary>
        /// Used for searching inside of files.
        /// </summary>
        /// <param name="needle">What to find.</param>
        /// <param name="path">The file to find the needle in.</param>
        /// <param name="buffersize">Size of the memory buffer to use in bytes (4096 is default).</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>The offsets of all instances of "needle."</returns>
        public static int[] Find(byte[] needle, string path, int buffersize, int limit)
        {
            FileStream fs = new FileStream(path, FileMode.Open);
            BinaryReader br = new BinaryReader(fs);
            FileInfo fi = new FileInfo(path);
            List<int> found = new List<int>();
            int counter = 0;
            int offset = -1;
            int overallposition = 0;
            bool searching = false;
            bool exitwhile = false;
            while (fs.Position != fi.Length)
            {
                // Get data into the buffer.
                byte[] haystack;
                if (buffersize + fs.Position > fi.Length)
                {
                    haystack = br.ReadBytes(Convert.ToInt32(fi.Length - fs.Position));
                }
                else if (buffersize < 0)
                {
                    haystack = br.ReadBytes(Convert.ToInt32(fi.Length));
                }
                else
                {
                    haystack = br.ReadBytes(buffersize);
                }

                for (int h = 0; h < haystack.Length; h++)
                {
                    if (needle[counter] != haystack[h])
                    {
                        if (searching)
                        {
                            if (Math.Sign(h - counter) == -1)
                            {
                                overallposition -= counter;
                                fs.Position = overallposition;
                                overallposition -= buffersize;
                            }
                            else
                            {
                                h -= counter;
                            }
                            counter = 0;
                            offset = -1;
                            searching = false;
                        }
                    }
                    else
                    {
                        if (offset == -1)
                        {
                            searching = true;
                            offset = overallposition + h;
                        }
                        if (counter == (needle.Length - 1))
                        {
                            if ((limit != found.Count) || (limit == -1))
                            {
                                found.Add(offset);
                            }
                            else
                            {
                                exitwhile = true;
                                break;
                            }
                            counter = 0;
                            offset = -1;
                        }
                        else
                        {
                            counter++;
                        }
                    }
                }
                if (exitwhile)
                {
                    break;
                }
                overallposition += buffersize;
            }
            br.Close();
            fs.Close();

            // Copy the data from List<> to int[] and return it.
            int[] output = new int[found.Count];
            for (int i = 0; i < output.Length; i++)
            {
                output[i] = found[i];
            }
            return output;
        }
        /// <summary>
        /// Used for searching inside a byte array (small files/strings).  Do not use this for files unless you are going to be performing many searches in rapid succession.
        /// </summary>
        /// <param name="needle">What to find.</param>
        /// <param name="haystack">The array to find the needle in.</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>The offsets of all instances of "needle."</returns>
        public static int[] Find(byte[] needle, byte[] haystack, int limit)
        {
            List<int> found = new List<int>();
            int counter = 0;
            int offset = -1;
            bool searching = false;
            for (int h = 0; h < haystack.Length; h++)
            {
                if (needle[counter] != haystack[h])
                {
                    if (searching)
                    {
                        h = h - counter;
                        counter = 0;
                        offset = -1;
                        searching = false;
                    }
                }
                else
                {
                    if (offset == -1)
                    {
                        searching = true;
                        offset = h;
                    }
                    if (counter == (needle.Length - 1))
                    {
                        if ((limit != found.Count) || (limit == -1))
                        {
                            found.Add(offset);
                        }
                        else
                        {
                            break;
                        }
                        counter = 0;
                        offset = -1;
                    }
                    else
                    {
                        counter++;
                    }
                }
            }

            int[] output = new int[found.Count];
            for (int i = 0; i < output.Length; i++)
            {
                output[i] = found[i];
            }
            return output;
        }
        /// <summary>
        /// Searches a single dimension byte array both ways.  For instance, searching for "racecar" in "racecar" will return two instances.
        /// </summary>
        /// <param name="needle">What to find.</param>
        /// <param name="haystack">The array to find the needle in.</param>
        /// <returns>An array of all the instances found.</returns>
        public static Found1D[] Find1D(byte[] needle, byte[] haystack)
        {
            List<Found1D> found = new List<Found1D>();

            for (int x = 0; x < haystack.Length; x++)
            {
                if (needle[0] == haystack[x])
                {
                    bool complete = true;

                    #region " Right "
                    try
                    {
                        complete = true;
                        for (int n = needle.Length - 1; n > 0; n--)
                        {
                            if (needle[n] != haystack[x + n])
                            {
                                complete = false;
                                break;
                            }
                        }
                    }
                    catch
                    {
                        complete = false;
                    }
                    finally
                    {
                        if (complete)
                        {
                            found.Add(new Found1D(x, x + needle.Length - 1));
                        }
                    }
                    #endregion

                    #region " Left "
                    try
                    {
                        complete = true;
                        for (int n = needle.Length - 1; n > 0; n--)
                        {
                            if (needle[n] != haystack[x - n])
                            {
                                complete = false;
                                break;
                            }
                        }
                    }
                    catch
                    {
                        complete = false;
                    }
                    finally
                    {
                        if (complete)
                        {
                            found.Add(new Found1D(x, x - needle.Length + 1));
                        }
                    }
                    #endregion
                }
            }

            Found1D[] output = new Found1D[found.Count];
            for (int i = 0; i < output.Length; i++)
            {
                output[i] = found[i];
            }
            return output;
        }
        /// <summary>
        /// Searches a two dimensional byte array in all directions.  This works great for solving word searches. ;)
        /// </summary>
        /// <param name="needle">What to find.</param>
        /// <param name="haystack">The array to find the needle in.</param>
        /// <returns>An array of all the instances found.</returns>
        public static Found2D[] Find2D(byte[] needle, byte[,] haystack)
        {
            List<Found2D> found = new List<Found2D>();

            for (int y = 0; y <= haystack.GetUpperBound(0); y++)
            {
                for (int x = 0; x < haystack.GetUpperBound(1); x++)
                {
                    if (needle[0] == haystack[y, x])
                    {
                        bool complete = true;

                        #region " Bottom Center "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y + n, x])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x, y + needle.Length - 1)));
                            }
                        }
                        #endregion

                        #region " Bottom Left "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y + n, x - n])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x - needle.Length + 1, y + needle.Length - 1)));
                            }
                        }
                        #endregion

                        #region " Middle Left "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y, x - n])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x - needle.Length + 1, y)));
                            }
                        }
                        #endregion

                        #region " Top Left "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y - n, x - n])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x - needle.Length + 1, y - needle.Length + 1)));
                            }
                        }
                        #endregion

                        #region " Top Center "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y - n, x])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x, y - needle.Length + 1)));
                            }
                        }
                        #endregion

                        #region " Top Right "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y - n, x + n])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x + needle.Length - 1, y - needle.Length + 1)));
                            }
                        }
                        #endregion

                        #region " Middle Right "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y, x + n])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x + needle.Length - 1, y)));
                            }
                        }
                        #endregion

                        #region " Bottom Right "
                        try
                        {
                            complete = true;
                            for (int n = needle.Length - 1; n > 0; n--)
                            {
                                if (needle[n] != haystack[y + n, x + n])
                                {
                                    complete = false;
                                    break;
                                }
                            }
                        }
                        catch
                        {
                            complete = false;
                        }
                        finally
                        {
                            if (complete)
                            {
                                found.Add(new Found2D(new Point(x, y), new Point(x + needle.Length - 1, y + needle.Length - 1)));
                            }
                        }
                        #endregion
                    }
                }
            }

            Found2D[] output = new Found2D[found.Count];
            for (int i = 0; i < output.Length; i++)
            {
                output[i] = found[i];
            }
            return output;
        }
        /// <summary>
        /// Searches a three dimensional byte array in all directions.  This could have a lot of potential in the field of cryptography.
        /// </summary>
        /// <param name="needle">What to find.</param>
        /// <param name="haystack">The array to find the needle in.</param>
        /// <returns>An array of all the instances found.</returns>
        public static Found3D[] Find3D(byte[] needle, byte[,,] haystack)
        {
            List<Found3D> found = new List<Found3D>();

            for (int z = 0; z <= haystack.GetUpperBound(2); z++)
            {
                for (int y = 0; y <= haystack.GetUpperBound(1); y++)
                {
                    for (int x = 0; x < haystack.GetUpperBound(0); x++)
                    {
                        if (needle[0] == haystack[z, y, x])
                        {
                            bool complete = true;

                            #region " Top Layer "
                            #region " Bottom Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y + n, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y + needle.Length - 1, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Bottom Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y + n, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y + needle.Length - 1, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Bottom Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y + n, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y + needle.Length - 1, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Middle Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Middle Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Middle Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Top Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y - n, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y - needle.Length + 1, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Top Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y - n, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y - needle.Length + 1, z + needle.Length - 1)));
                                }
                            }
                            #endregion

                            #region " Top Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z + n, y - n, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y - needle.Length + 1, z + needle.Length - 1)));
                                }
                            }
                            #endregion
                            #endregion

                            #region " Middle Layer "
                            #region " Bottom Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y + n, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y + needle.Length - 1, z)));
                                }
                            }
                            #endregion

                            #region " Bottom Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y + n, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y + needle.Length - 1, z)));
                                }
                            }
                            #endregion

                            #region " Middle Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y, z)));
                                }
                            }
                            #endregion

                            #region " Top Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y - n, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y - needle.Length + 1, z)));
                                }
                            }
                            #endregion

                            #region " Top Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y - n, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y - needle.Length + 1, z)));
                                }
                            }
                            #endregion

                            #region " Top Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y - n, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y - needle.Length + 1, z)));
                                }
                            }
                            #endregion

                            #region " Middle Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y, z)));
                                }
                            }
                            #endregion

                            #region " Bottom Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z, y + n, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y + needle.Length - 1, z)));
                                }
                            }
                            #endregion
                            #endregion

                            #region " Bottom Layer "
                            #region " Bottom Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y + n, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y + needle.Length - 1, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Bottom Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y + n, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y + needle.Length - 1, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Bottom Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y + n, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y + needle.Length - 1, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Middle Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Middle Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Middle Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Top Left "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y - n, x - n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x - needle.Length + 1, y - needle.Length + 1, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Top Center "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y - n, x])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x, y - needle.Length + 1, z - needle.Length + 1)));
                                }
                            }
                            #endregion

                            #region " Top Right "
                            try
                            {
                                complete = true;
                                for (int n = needle.Length - 1; n > 0; n--)
                                {
                                    if (needle[n] != haystack[z - n, y - n, x + n])
                                    {
                                        complete = false;
                                        break;
                                    }
                                }
                            }
                            catch
                            {
                                complete = false;
                            }
                            finally
                            {
                                if (complete)
                                {
                                    found.Add(new Found3D(new Point3D(x, y, z), new Point3D(x + needle.Length - 1, y - needle.Length + 1, z - needle.Length + 1)));
                                }
                            }
                            #endregion
                            #endregion
                        }
                    }
                }
            }

            Found3D[] output = new Found3D[found.Count];
            for (int i = 0; i < output.Length; i++)
            {
                output[i] = found[i];
            }
            return output;
        }
    }
}
