using System;
using System.IO;
using System.Text;


/*
 * Copyright 2007 Lee Glasser
 * 
 * This code is distributed under the terms of the GNU General Public License
 * (GPL).
 * 
 * This file is part of the Includes package.
 * 
 * "Includes.FindReplace" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 * 
 * "Includes.FindReplace" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with "Includes.FindReplace."  If not, see <http://www.gnu.org/licenses/>.
 */


namespace FindReplace
{
    /// <summary>
    /// This class is for searching for byte sequences (case sensitive) in small files (the smaller the better).  It stores the entire file in the memory.
    /// </summary>
    public class SmallFile
    {
        private byte[] _Buffer;

        /// <summary>
        /// Creates a new instance of SmallFile where the source (path) is used for all subsequent methods.
        /// </summary>
        /// <param name="path">The source file.</param>
        public SmallFile(string path)
        {
            Open(path);
        }

        /// <summary>
        /// Opens a new file.
        /// </summary>
        /// <param name="path">The path to file.</param>
        /// <returns>True on success, false on failure.</returns>
        public bool Open(string path)
        {
            try
            {
                FileInfo fi = new FileInfo(path);
                FileStream fs = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.Read);
                BinaryReader br = new BinaryReader(fs);
                _Buffer = br.ReadBytes(Convert.ToInt32(fi.Length));
                br.Close();
                fs.Close();
                return true;
            }
            catch
            {
                return false;
            }
        }
        /// <summary>
        /// Dumps the buffer to a file at "path."
        /// </summary>
        /// <param name="path">The path to file.</param>
        /// <returns>True on success, false on failure.</returns>
        public bool Save(string path)
        {
            try
            {
                FileStream fs = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.None);
                BinaryWriter bw = new BinaryWriter(fs);
                bw.Write(_Buffer);
                bw.Flush();
                bw.Close();
                fs.Close();
                return true;
            }
            catch
            {
                return false;
            }
        }
        /// <summary>
        /// Simply gets all instances of "find" in the file.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <returns>The offsets of all instances of "find."</returns>
        public int[] Find(string find)
        {
            return Search.Find(Encoding.ASCII.GetBytes(find), _Buffer, -1);
        }
        /// <summary>
        /// Gets the "limit" of instances of "find" in the file.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>The offsets of all instances of "find."</returns>
        public int[] Find(string find, int limit)
        {
            return Search.Find(Encoding.ASCII.GetBytes(find), _Buffer, limit);
        }
        /// <summary>
        /// Finds "find" and replaces it with "replace" in the buffer.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(string find, string replace)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), -1);
        }
        /// <summary>
        /// Finds "find" and replaces it with "replace" in the buffer.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(string find, string replace, int limit)
        {
            return FindReplace(Encoding.ASCII.GetBytes(find), Encoding.ASCII.GetBytes(replace), limit);
        }
        /// <summary>
        /// Finds "find" and replaces it with "replace" in the buffer.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(byte[] find, byte[] replace)
        {
            return FindReplace(find, replace, -1);
        }
        /// <summary>
        /// Finds "find" and replaces it with "replace" in the buffer.
        /// </summary>
        /// <param name="find">What to find.</param>
        /// <param name="replace">What to replace found with.</param>
        /// <param name="limit">The maximum to return (-1 is default).</param>
        /// <returns>Total number replaced.</returns>
        public int FindReplace(byte[] find, byte[] replace, int limit)
        {
            int[] offsets = Search.Find(find, _Buffer, limit);
            int i;
            for (i = 0; i < offsets.Length; i++)
            {
                for (int j = 0; j < replace.Length; i++)
                {
                    _Buffer[offsets[i] + j] = replace[j];
                }
            }
            return i;
        }
    }
}
